/****************************************************************************
**
** Copyright (C) 2007-2007 Trolltech ASA. All rights reserved.
**
** This file is part of the QtXMLPatterns module of the Qt Toolkit.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  Please review the following information to ensure GNU
** General Public Licensing requirements will be met:
** http://trolltech.com/products/qt/licenses/licensing/opensource/
**
** If you are unsure which license is appropriate for your use, please
** review the following information:
** http://trolltech.com/products/qt/licenses/licensing/licensingoverview
** or contact the sales department at sales@trolltech.com.
**
** In addition, as a special exception, Trolltech gives you certain
** additional rights. These rights are described in the Trolltech GPL
** Exception version 1.0, which can be found at
** http://www.trolltech.com/products/qt/gplexception/ and in the file
** GPL_EXCEPTION.txt in this package.
**
** In addition, as a special exception, Trolltech, as the sole copyright
** holder for Qt Designer, grants users of the Qt/Eclipse Integration
** plug-in the right for the Qt/Eclipse Integration to link to
** functionality provided by Qt Designer and its related libraries.
**
** Trolltech reserves all rights not expressly granted herein.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qxmlresultitems_p.h"
#include "qitem_p.h"

QT_BEGIN_NAMESPACE

/*!
  \class QXmlResultItems
  \brief The QXmlResultItems class represents the items that are the result of an XQuery query.
  \reentrant
  \since 4.4
  \ingroup xml-tools

  QXmlResultItems represent the items of a XQuery query, and allows them
  to be iterated like a forward iterator.

  \code
  QXmlQuery query;
  query.setQuery("<e/>, 1, 'two'");
  QXmlResultItems result;

  if(query.isValid())
  {
    query.evaluateToResult(&result);
    QXmlItem item(result.next())
    while(!item.isNull())
    {
        // Use item
        item = result.next();
    }

    if(result.hasError())
        return "Runtime error!";
  }
  \endcode

  QXmlResultItems lazily evaluates the query as it is executed. One effect of this
  is that an error can occur. In that case hasError() will return \c true, and next()
  and current() returns a \c null QXmlItem.

  QXmlResultItems can be seen as an iterator that advances over the sequence of items
  that a query produced. Each call to next() advances the QXmlResultItems instance to
  "point" to the next item in the sequence and returns it, and current() returns the
  item that next() returned the last time it was called.

  \sa QXmlItem::isNode(), QXmlItem::isAtomicValue(), QXmlNodeModelIndex
 */

/*!
 Constructs a QXmlResultItems instance.
 */
QXmlResultItems::QXmlResultItems() : d_ptr(new QXmlResultItemsPrivate())
{
}

/*!
 Destructs this QXmlResultItems instance.
 */
QXmlResultItems::~QXmlResultItems()
{
    delete d_ptr;
}

/*!
 
  On a default constructed QXmlResultItems instance, a \c null QXmlItem is returned.
 */
QXmlItem QXmlResultItems::next()
{
    Q_D(QXmlResultItems);
    if(d->hasError)
        return QXmlItem();

    try
    {
        d->current = QPatternist::Item::toPublic(d->iterator->next());
        return d->current;
    }
    catch(bool)
    {
        d->current = QXmlItem();
        return QXmlItem();
    }
}

/*!
  Returns the c
 
  On a default constructed QXmlResultItems instance, a \c null QXmlItem is returned.
 */
QXmlItem QXmlResultItems::current() const
{
    Q_D(const QXmlResultItems);

    if(d->hasError)
        return QXmlItem();
    else
        return d->current;
}

/*!
 If an error occurs during the evaluation, this function returns \c true.

 By default \c false is returned.
 */
bool QXmlResultItems::hasError() const
{
    Q_D(const QXmlResultItems);
    return d->hasError;
}

QT_END_NAMESPACE

